/***************************************************************

	Windows version of Fishnet( Ported from Linux Version )
	Author: Rajesh Hegde K ( krhegde@cs.washington.edu )
	April 21, 2002

***************************************************************/

#include <sys/types.h>
#include <winsock2.h>
#include <ws2tcpip.h>
#include <stdlib.h>
#include <stdio.h>

#include <errno.h>
//#define __USE_GNU
#include <string.h>
#include <stdlib.h>


#include <time.h>
#include <ctype.h>
#include <assert.h>
#include <fcntl.h>
#include <conio.h>
#include <windows.h>
#include <process.h>
#include "winfishguts.h"
#include "winfish.h"

#define MAX_NEIGHBORS 12
#define MAXPACKET 1024
#define MINIMUM_TIME_BETWEEN_TWO_SENDS .01



#define MIN(a,b) (((a)>(b)) ? (b) : (a))
#define MAX(a,b) (((a)>(b)) ? (a) : (b))
#define InRange(low, val, up) (((low) <= (val)) && ((val) <= (up)))
#define PRINT_AND_BAIL(x) do { fish_debugframe(FISH_DEBUG_NONE, "Error", \
		                               pkt, frame_length); \
                               errorabort(x); } while(0);

/*@-booltype boolean@*/



static char *fishhead_host = NULL;
static int fishhead_port = -1;
static int self_address = -1;
static char self_domain[8];

static boolean init = FALSE;
/* the sockets are -1 so that join can test whether they're
   valid to protect against being called twice */
static int sockfishhead = -1;
static int sockfishbits = -1;
static struct sockaddr_in headaddr;
static struct sockaddr_in bitsaddr;
static time_t last_send_time;  

/* +1 to allow null termination */
static struct neighbor neighbors[MAX_NEIGHBORS+1];

static receive_handler recvhandler;
static keyboard_handler keybhandler;

static void read_neighbor_list(int sd_fishhead); 
static int find_by_sock_address(const struct sockaddr_in addr);
static void frame_verification(const char *frame, int len);

static int fish_constructmask(fd_set *prd);
static void fish_processbindings(fd_set *prd);

/* neighbor history routines; don't fail on a send
   frame if the neighbor was there before. */
#define MAX_NEIGHBOR_HIST 1024
static int neighbor_hist[MAX_NEIGHBOR_HIST];


static int find_in_neighbor_history(int naddr) 
{
  int i; 
  for(i=0; i<MAX_NEIGHBOR_HIST && neighbor_hist[i]!= 0  && neighbor_hist[i]!= naddr ;   i++);

  if(neighbor_hist[i] == naddr) 
  {
	return i;
  }
  else
  {
    return -1;
  }
}

static void insert_neighbor_history(int naddr) 
{
  int i;
  for(i=0; i<MAX_NEIGHBOR_HIST && neighbor_hist[i]!= 0 && neighbor_hist[i]!= naddr ;  i++);

  if(i < MAX_NEIGHBOR_HIST) 
  {
    if(neighbor_hist[i] == 0) 
	{
		neighbor_hist[i] = naddr;
	}
  } 
  else 
  {
    printf("Non-fatal error: ran out of neighbor history space.\n");
  }
}

static void update_neighbor_history(void) 
{
  int i;
  
  for (i=0; i< MAX_NEIGHBORS && neighbors[i].address != 0 ; i++);

  insert_neighbor_history(neighbors[i].address);
}

/* used when the likely cause is system (possibly TA) error */
static void errorexit(const char* msg, int syserr) 
{
  fprintf(stderr, "%s\n",msg);
  if (syserr != 0) perror("\n fish ");
  exit(EXIT_FAILURE);
}

/* used when the likely cause is programmer (read: student) error */
extern void errorabort(const char* msg) 
{
  fprintf(	stderr, "** %s **\n", msg);
  fprintf(	stderr,	"Generating core dump.  Use: \n"
					"  gdb [programname] core\n"
					"and use the commands: where, up, down, list, and print,\n"
					"if the error messages above (in **'s) haven't helped.\n");

  /* bail, generating core dump (if not limited, see getrlimit(2)) */
  abort();
}

static void fish_init(void) 
{
  (void)time(&last_send_time); 
  init = TRUE;
  srand((unsigned int)time(0));
  return;
}

/* delivers a message */
static void deliver(int fromid, char *buf, int msglen) 
{

 char msg[32];
 sprintf(msg, "fish: received frame from %d", fromid);
 fish_debugframe(FISH_DEBUG_ALL, msg, buf, msglen);
 if (recvhandler != NULL)
 {
    (*recvhandler)(fromid, buf, msglen);
 }
}


/* handles input from other fishnet nodes */
static void receive_handler_wrapper(int sd_fishbits ) 
{
  char buf[1024];
  int msglen;
  struct sockaddr_in from;
  socklen_t fromlen = sizeof(struct sockaddr_in);
  int fromid;
  
  memset(&buf, 0, sizeof(buf));
  msglen = recvfrom(sockfishbits, buf, sizeof(buf), 0, (struct sockaddr*)&from, &fromlen);
  if( msglen == SOCKET_ERROR  || msglen == 0)
  {
	  fish_debug(FISH_DEBUG_ALL, "ERROR: recvfrom() failed: %d\n", WSAGetLastError() );
	  return;
  }

  fromid = find_by_sock_address(from);
  if (fromid < 1) 
  {
    fish_debug(FISH_DEBUG_NONE, "Received packet with entry for socket address!\n");
  }
  else 
  {
    deliver(fromid, buf, msglen);
  }
}


extern void fish_recvhook(receive_handler fn) 
{
  if (!init) 
  {
	  fish_init();
  }
  
  recvhandler = fn;
}

extern void keyboard_handler_wrapper(int stdin_fd) 
{
  char buf[1024];
  
  fgets(buf,sizeof(buf),stdin);
  if(feof(stdin)) 
  {
    fish_debug(FISH_DEBUG_NONE, "fish: stdin closed, type 'exit' instead of ^D, or hit ^C now to quit.\n");
    fish_remove_readhook(stdin_fd);
    /* exit(EXIT_SUCCESS); */
    return;
  }

  if(strncmp(buf, "exit", 4)==0) 
  {
    fish_debug(FISH_DEBUG_NONE, "fish: exiting\n");
    exit(EXIT_SUCCESS); 
  }
  
  fish_debug(FISH_DEBUG_ALL, "fish: received keyboard input\n");
  if (keybhandler != NULL)
  {
    (*keybhandler)(buf);
  }
}

extern void fish_keybhook(keyboard_handler fn) 
{
  if (!init) 
  {
	  fish_init();
  }
  
  fish_readhook( fileno( stdin ), keyboard_handler_wrapper);
  
  keybhandler = fn;
}

/* sends a frame to a udp port, ip address */

static int udpsend(struct sockaddr_in to, int naddr,  void* frame, int len) 
{
  int ok;
  ok = sendto( sockfishbits, (const char*)frame, len, 0, (struct sockaddr*)&to, sizeof(to)); 

  if (ok < 0) 
  {
    if (errno == WSAECONNREFUSED  ) 
	{
      fish_debug(FISH_DEBUG_TOPOLOGY,  "fish: warning - neighbor %d seems to have failed\n", naddr);
    } 
	else 
	{
      fish_debug(FISH_DEBUG_NONE, "fish: while sending to %d at %s:%d\n", naddr, inet_ntoa(to.sin_addr), htons(to.sin_port));
      fish_debug(FISH_DEBUG_NONE,  "fish: experienced error %d: %s\n", errno, strerror(errno));
      errorexit("fish: error, can't send\n", 1);
    }
  } 

  // limit sending rate
  Sleep(10);

  return ok;
}
/* 
 * Actually send a frame to a neighbor; called by fish_send 
 * The result is <0 on failure.
 */
static int realsend(int n, int naddr, void* frame, int len) 
{
  struct sockaddr_in to;

  if ((unsigned int)((rand() >> 16 | rand()) % 1000)  < (unsigned int)neighbors[n].lossrate_x_1000) 
  {
    fish_debug(FISH_DEBUG_ALL, "fish: Dropped frame from %d to %d\n", self_address, naddr );
    /* should be a silent failure. */
    return 0;
  }

  memset((void *)&to, 0, sizeof(to));
  to.sin_family      = AF_INET;
  to.sin_addr        = neighbors[n].node_ip_address;
  to.sin_port        = neighbors[n].node_udp_port;

  return udpsend(to,naddr,frame,len);
}

extern int fish_send(int naddr, void* frame, int len) 
{
  int n;
  int i;

  struct sockaddr_in to;

  int status = -1;

  if (!init) 
  {
	  fish_init();
  }

  if(sockfishbits < 0) 
  {
    errorabort("fish: must join the network before sending\n");
  }

  if (len < 1) 
  {
    errorabort("fish: zero-length frame\n");
  }

  if (len > MAXPACKET) 
  {
    errorabort("fish: frame too long\n");
  }

  frame_verification( (const char*)frame,len);

  switch (naddr) 
  {

  case ANY_NEIGHBOR:
    // anycast
    for (i=0; (i < MAX_NEIGHBORS) && (neighbors[i].address != 0); i++);

    if (i==0) 
	{
      fish_debug(FISH_DEBUG_NONE, "fish: no neighbors to send to in ANYCAST\n");
      return -1;
    } 
	else 
	{
      status = realsend(rand() % i, naddr, frame, len);
    }
    break;
  
  case ALL_NEIGHBORS:
    // broadcast
    for (i=0; (i < MAX_NEIGHBORS) && (neighbors[i].address != 0); i++) 
	{
      status = realsend(i, naddr, frame, len);
    }
    break;

   
    default:
      
      // local loopback delivery
      if (naddr == self_address) 
	  {
		memset((void *)&to, 0, sizeof(to));
		to.sin_family      = AF_INET;
		to.sin_addr.s_addr = inet_addr("127.0.0.1");
		to.sin_port        = bitsaddr.sin_port ;
  		status = udpsend(to,self_address,frame,len);
      } 
	  else 
	  {
		// neighborcast
		n = -1;
		for (i = 0; i < MAX_NEIGHBORS; i++) 
		{
			if (neighbors[i].address == naddr) 
			{
				n = i; 
			}
		}
		if (n < 0) 
		{
			/* only abort if this neighbor never existed. */
			if(find_in_neighbor_history(naddr) == -1) 
			{
				fish_debug(FISH_DEBUG_NONE,"fish: %d not a valid neighbor\n", naddr);
				return -1;

			} 
			else 
			{
				errorabort("fish: not a valid neighbor\n"); 
			}
		} 

		status = realsend(n, naddr, frame, len);
      }
  }

  if (status >= 0) 
  {
    fish_debugframe(FISH_DEBUG_ALL, "fish: Sent frame", frame, len);
    return 0;
  }

  return -1;
}


void parse_location(const char* location, char** host, int* port) 
{

  char* separator = strchr(location, ':');
  long hostlength = separator - location;
  if (separator == NULL) 
  {
    errorabort("fish: invalid port number\n");
  }

  *host = (char*)malloc(hostlength + 1);
  memset( *host, 0, hostlength + 1 );
  strncpy(*host, location, hostlength);
  //*host[ hostlength] = '\0';

  *port = atoi(separator+1);
}

extern void fish_joinnetwork(	const char* fishhead_location,
					            const char* domain,
								int addr) 
{
  int ok;
  socklen_t address_length_buf;

  struct hostent *hp;
  struct in_addr *firstAddress;
  char	 messagebuffer[1024];
  struct fish_message *pfm = (struct fish_message *)messagebuffer;
  struct joinrequest *jreq = &pfm->body.joinreq; 

  if (sockfishhead > 0 || sockfishbits > 0) 
  {
    errorabort("fish: error, already joined\n");
  }

  if ( ! InRange(1, addr, MAX_NODE_ADDRESS)) 
  {
    errorexit("fish: not a valid fishnet address", 0);
  }

  if (!init) 
  {
	  fish_init();
  }

  // parse the fishhead location
  parse_location( fishhead_location, &fishhead_host, &fishhead_port );

  self_address = addr;
  if (domain != NULL) 
  {
    strncpy(self_domain, domain, 8);
  } 
  else 
  {
    self_domain[0] = '\0';
  }

  // Open a UDP Port
  sockfishbits = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
  if (sockfishbits < 0) 
  {
    errorexit("fish: error,  can't get fishbits socket\n", 1);
  }

  bzero((char *) &bitsaddr, sizeof(bitsaddr));
  bitsaddr.sin_family      = AF_INET;
  bitsaddr.sin_addr.s_addr = htonl(INADDR_ANY);
  bitsaddr.sin_port        = htons(0);

  ok = bind(sockfishbits, (struct sockaddr*)&bitsaddr, sizeof(bitsaddr));
  if (ok < 0) 
  {
    errorexit("fish: error, can't bind fishbits socket\n", 1);
  }

  // Bind the socket to a handler off the main select() loop
  fish_readhook(sockfishbits, receive_handler_wrapper);

  // Get the port number allocated by the system, so that 
  // we can tell fishhead how others can contact us.
  address_length_buf = sizeof(bitsaddr);
  ok = getsockname(	sockfishbits,
					(struct sockaddr *)&bitsaddr, 
					&address_length_buf);
  if (ok < 0) 
  {
    errorexit("fish: error, can't find port number of fishbits socket\n", 1);
  }

  /// Code handling the network manager
  sockfishhead = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
  if (sockfishhead < 0) 
  {
    errorexit("fish: error, can't get fishhead socket\n", 0);
  }

  // Convert hostname to  ipaddress
  if ( (hp = gethostbyname(fishhead_host)) == NULL) 
  {
	errorexit(" Failure in mapping host name to IP",0);
  }
  
  firstAddress = (struct in_addr *) hp->h_addr_list[0];
  
  if (firstAddress == NULL) 
  {
    errorexit(" Failure in mapping host name to IP",0);
  }
  
  bzero((char *) &headaddr, sizeof(headaddr));
  memcpy(&headaddr.sin_addr, firstAddress, sizeof(struct in_addr));
  headaddr.sin_family      = AF_INET;
  headaddr.sin_port        = htons(fishhead_port);
  
  // Connect to the network manager
  ok = connect(	sockfishhead,
				(struct sockaddr*) &headaddr,
				sizeof(headaddr));
  if (ok < 0)
  {
    errorexit("fish: error, can't get fishhead connection\n", 1);
  }

  // Send a join request to the network manager
  pfm->type   = FISH_MSG_TYPE_JOINREQUEST;
  pfm->length = MIN_MESSAGE_SIZE + (sizeof(struct joinrequest));
  jreq->node_address = addr;
  jreq->node_udp_port = bitsaddr.sin_port;
  strncpy(jreq->username, getenv("USER") == NULL ? "USER" : getenv( "USER") , 8);
  strncpy(jreq->domain, self_domain, 8);
  
  int bytesent = send(sockfishhead, (char*)messagebuffer, (size_t)pfm->length, 0 );
 
  /* Now wait to hear the return message from the server
   * telling you your neighbor list... */
  (void)read_neighbor_list(sockfishhead);
  printneighbors(neighbors);
  
  /* setup input handler to get neighbor list updates from fishhead */
  fish_readhook(sockfishhead, read_neighbor_list);

  fish_debug(FISH_DEBUG_TOPOLOGY, "fish: joined network %s:%i as %s:%i\n", fishhead_host, fishhead_port, self_domain, self_address);
}

extern const char* fish_getheadhost(void) 
{
  return fishhead_host;	  
}

extern int fish_getheadport(void) 
{
  return fishhead_port;
}

extern int fish_getaddress(void) 
{
  return self_address;
}

extern const char* fish_getdomain(void) 
{
  return self_domain;
}

extern void fish_main(void) 
{
 DWORD ithreadid;
  
  if (!init) 
  {
	  fish_init();
  }

  if(sockfishbits<0) 
  {
    errorabort("fish: must join the network before calling fish_main\n");
  }

  fish_debug(7, "fish: in main\n");

  CreateThread( NULL, 0, (LPTHREAD_START_ROUTINE)kbrdthread, NULL, 0,  &ithreadid );


  
  while (1) 
  {
    fd_set rd;
	memset( &rd, 0, sizeof( fd_set ));
    int max = 0;
    struct timeval wait;
    int ok;

    max = fish_constructmask(&rd);

 //   /* patch for apparent problem with stdin.  I don't
 //      understand why this is necessary, but fixing it this
 //      way is better than leaving it broken. */
    
	//if(FD_ISSET(fileno(stdin), &rd) ) 
	//{
      fish_remove_readhook(fileno(stdin));
      max = fish_constructmask(&rd);
	  //if( _kbhit() && ithreadid == -1 )	  
	  //{		  	
	//	  keyboard_handler_wrapper( fileno(stdin) );
	  //}
    //}

    ok = select(max+1, &rd, 0, 0, fish_time_to_next_event(&wait) ? &wait : 0);

	//
	// select has a weird behaviour on windows.. So not checking the error.. I know it is ineffient
	// but will do for the projects - rajesh
	//
	//continue;
    if (ok < 0) 
	{
      if(errno != EINTR) 
	  {
        if(errno == EBADF) 
		{
          int i;
          printf("Bad file descriptor error: be careful to call fish_remove_readhook() with close()\n");
         
		  for(i=0; i<max; i++) 
		  {
            if(FD_ISSET(i, &rd)  < 0) 
			{
              printf("  file descriptor #%d is suspicious: %s\n", 
                     i, strerror(errno));
            }
          }
          
        } 
		else 
		{
          perror("select");
        }
        
		errorabort("fish: error calling select()");
      } 
	  else 
	  {
        continue;
      }
    }

    /* do scheduled things */
    fish_executepending();

    /* handle anything that uses sockets, including
       fishhead, fishbits, and keyboard */
    fish_processbindings(&rd);
	continue;
  }
}



static int find_by_sock_address(const struct sockaddr_in addr) 
{
  int i ;
  
  /* First check if this is a local packet */
  if ((inet_addr("127.0.0.1") == addr.sin_addr.s_addr) && (bitsaddr.sin_port == addr.sin_port))
  {
    return self_address;
  }
  
  for(i=0; i< MAX_NEIGHBORS; i++) 
  {
    if ((neighbors[i].node_ip_address.s_addr == addr.sin_addr.s_addr) && (neighbors[i].node_udp_port == addr.sin_port)) 
	{
      return neighbors[i].address;
    }
  }
  return -1;
}

/* helper function for frame_verification (below) that 
   verifies that the length of a text containing frame
   (either assignment 1 style or ECHO/ECHO_RESPONSE is no
   larger than necessary */  
static boolean checkStringLength(	const char *string, 
									unsigned int string_length) 
{
  char* szTemp = (char*)malloc( string_length + 1 );
  memcpy( szTemp, string, string_length );
  szTemp[ string_length ] = '\0';
	/* strnlen is a gnu-specific function, easily re-implemented */
  if(string_length != 0 && strlen(szTemp ) < string_length-1 ) 
  {
    fprintf(stderr, "The frame containing the following text:\n"
            "%*s\n"
            "was sent with a length of %u bytes, "
            "but only requires %d bytes\n", 
            string_length, szTemp, 
            string_length, strlen(szTemp));

    free( szTemp );
	return(FALSE);
  } 
  else
  {
	free( szTemp );	
	return(TRUE);
  }
}

/* abort()s on failure */
static void 
frame_verification(const char *frame, const int frame_length) 
{
  int i = 0;
  int textframe = 0;
  int toCheckLen = 0;
  const struct packet *pkt = (const struct packet *)frame;

  if(frame == NULL) 
  {
    errorabort("Null frame");
  }

  if (frame_length > MTU)
  {
    errorabort("Size of frame > MTU");

	if (frame_length < 0) 
	{
		errorabort("Frame has negative size");
	}
  
	// Check if the frame is text only...
	toCheckLen = MIN(16, frame_length);
	for(i=0, textframe = 1; i< toCheckLen; i++) 
	{
		textframe &= (isprint(frame[i]) || isspace(frame[i])) ;
	}
  }
  
  if (textframe) 
  {
    /* Verify that the frame length (the function parameter)
       is at most as long as the string plus a safe
       terminating zero.  This is not an egregious error,
       but verifying it here is probably the only way to
       make sure students pass a length parameter that is
       not too long. */
    if(!checkStringLength(frame, frame_length)) 
	{
      errorabort("Text frame was too long");
    }
    return;
  }
  
  // Packet is not just text
  if ( ! InRange(1, pkt->hdr.src, MAX_NODE_ADDRESS)) 
  {
    PRINT_AND_BAIL("In send: Invalid Source");
  }
  
  if ( ! InRange(-1, pkt->hdr.dst, MAX_NODE_ADDRESS)) 
  {
    PRINT_AND_BAIL("In send: Invalid Destination");
  }

  if ( (pkt->hdr.dst == ALL_NEIGHBORS) && (pkt->hdr.src != self_address)) 
  {
    PRINT_AND_BAIL("In send: src != self for packet sent to ALL_NEIGHBORS");
  }

  if ( ! InRange(1, pkt->hdr.ttl, MAX_TTL)) 
  {
    PRINT_AND_BAIL("In send: Invalid TTL");
  }

  if ( ! InRange(1, pkt->hdr.protocol, FISH_PROTOCOL_MAX)) 
  {
    PRINT_AND_BAIL("In send:Invalid Protcol");
  }
  
  // Protocols specific checkings...
  if (pkt->hdr.protocol == FISH_PROTOCOL_ROUTING)
  {
    if (pkt->hdr.ttl != 1)  
	{
      PRINT_AND_BAIL("In send: TTL must be 1 one for a Routing Packet");
    }
    if (pkt->hdr.src != self_address) 
	{
      PRINT_AND_BAIL("In send: Source for a Routing Packet must be same as self");
    }
  }

  
  if (pkt->hdr.protocol == FISH_PROTOCOL_TRANSPORT)
  {
    const struct transport_packet *tpkt =  (const struct transport_packet *)frame;

    if ( ! InRange(0,tpkt->transport_hdr.src_port,MAX_PORT_NUMBER)) 
	{
      PRINT_AND_BAIL("In send:Transport protocol packet has invalid src port");
    }
    if ( ! InRange(0,tpkt->transport_hdr.dst_port,MAX_PORT_NUMBER)) 
	{
      PRINT_AND_BAIL("In send:Transport protocol packet has invalid dst port");
    }
    
    if(!((tpkt->transport_hdr.fin==0) || (tpkt->transport_hdr.fin ==1))) 
	{
      PRINT_AND_BAIL("In send:Transport protocol packet has invalid fin (must be zero or one)");
    }
     
    if(!((tpkt->transport_hdr.syn==0) || (tpkt->transport_hdr.syn ==1)))
	{
      PRINT_AND_BAIL("In send:Transport protocol packet has invalid syn (must be zero or one)");
    }
  }
    
  if (pkt->hdr.protocol == FISH_PROTOCOL_ECHO_REQUEST  || pkt->hdr.protocol == FISH_PROTOCOL_ECHO_RESPONSE) 
  {
	if(!checkStringLength(pkt->data, frame_length-PACKET_HEADER_SIZE)) 
	{
		PRINT_AND_BAIL("In send: ECHO/ECHO_RESPONSE frame was too long");
	}
  }
  
}


static void read_neighbor_list(int sd_fishhead) 
{
#define MSG_BUF_SZ 1024

  char messagebuffer[MSG_BUF_SZ];
  int readlen;

  struct fish_message *pfm = (struct fish_message *)messagebuffer;

  bzero((char *) &messagebuffer, sizeof(messagebuffer));
  readlen = recv(sd_fishhead, (char*)messagebuffer, MSG_BUF_SZ, 0);
  if ( readlen == SOCKET_ERROR || readlen == 0 ) 
  {
	errorexit("Fishead died\n", 1);
  }

  fish_debug(FISH_DEBUG_TOPOLOGY, "Neighbor update received.\n");

  if(pfm->length > readlen) 
  {
    errorexit("read() failed, message not completely read",1);
  }
  
  if (pfm->type != FISH_MSG_TYPE_NEIGHBORLIST)
  {
    printmessage(pfm);
    errorabort("Above Message received from fish-head \n");
  }
  else 
  {
    int i;
    int num_neigh = (pfm->length - MIN_MESSAGE_SIZE) / sizeof(struct neighbor);
    // printmessage(pfm);
	
    for(i=0; i< num_neigh; i++) 
	{
      neighbors[i].address = pfm->body.neighbors[i].address ;
      neighbors[i].node_ip_address =  pfm->body.neighbors[i].node_ip_address ;
      neighbors[i].node_udp_port = pfm->body.neighbors[i].node_udp_port ;
      neighbors[i].lossrate_x_1000 = pfm->body.neighbors[i].lossrate_x_1000;
      assert(neighbors[i].lossrate_x_1000<=1000);
    }
    for(i=num_neigh; i< MAX_NEIGHBORS; i++) 
	{
      neighbors[i].address = 0;
    }
    update_neighbor_history();
  }
}

/* select bindings describe the handler to call
   when a socket becomes ready for writing */
static struct select_binding 
{
  struct select_binding *next;
  int sd;
  void (*handler)(int);
} *binding_list;

/* add a select binding */
extern void fish_readhook(int sd, void (*rdfn)(int)) 
{
  struct select_binding *psb =  (struct select_binding *)malloc(sizeof(struct select_binding));
  if( sd < 0) 
  {
    errorabort("bad socket descriptor passed to fish_readhook\n");
  }
  psb->sd = sd;
  psb->handler = rdfn;
  psb->next = binding_list;
  binding_list = psb;
}

/* remove a select binding */
extern void fish_remove_readhook(int sd) 
{
  struct select_binding *psb, *qsb;
  /* printf("removing hook for %d\n", sd); */
  for(psb = binding_list, qsb = NULL; psb != NULL && psb->sd != sd;   qsb=psb, psb=psb->next);

  if(psb != NULL) 
  { /* found it */
    if(qsb != NULL) 
	{ /* in the middle */
      qsb->next = psb->next;
    } 
	else 
	{ /* at the beginning */
      binding_list = psb->next;
    }
    free(psb);
  }
}

/* construct a select read mask from the select bindings */
static int fish_constructmask(fd_set *prd) 
{
  struct select_binding *psb;
  int max = 0;
  FD_ZERO(prd);
  for(psb = binding_list; psb != NULL; psb = psb->next) 
  {
    FD_SET(psb->sd, prd); 
    max = MAX(psb->sd, max);
  }
  return(max);
}

/* walk the bindings list and invoke the callbacks */
static void fish_processbindings(fd_set *prd) 
{
  struct select_binding *psb, *next;
  for(psb = binding_list; psb != NULL; psb = next) 
  {
    next = psb->next; /* grab this now, in case we remove
                         ourselves from the list */
    if(psb->sd >= 0 && psb->sd < 4096) 
	{
      if(FD_ISSET(psb->sd, prd)) 
	  {
        psb->handler(psb->sd);
        return; 
      }
    } 
	else 
	{
      printf("something got corrupted\n");
      exit(EXIT_FAILURE);
    }
  }
}

void SysTimeToTimeVal( SYSTEMTIME* psystime, timeval* ptimeval )
{
	ptimeval->tv_sec = ( ( psystime->wHour )* 60 + psystime->wMinute ) * 60 + psystime->wSecond;
	ptimeval->tv_usec = psystime->wMilliseconds * 1000;
}

DWORD __stdcall kbrdthread( void )
{
	while( true )
	{
		if( kbhit() )
		{
			keyboard_handler_wrapper( fileno(stdin) );
		}
	}
}