/***************************************************************

	Windows version of Fishnet( Ported from Linux Version )
	Author: Rajesh Hegde K ( krhegde@cs.washington.edu )
	April 21, 2002

***************************************************************/

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include "Winfish.h"
#include "Winfishguts.h"

struct eventlist 
{
  struct eventlist *next;
  struct timeval when;
  void (*event_handler)(void *);
  void *event_handler_argument;
};


void timersub( timeval * a, timeval* b, timeval* result) 
{

	do 
	{  
		(result)->tv_sec = (a)->tv_sec - (b)->tv_sec; 
		(result)->tv_usec = (a)->tv_usec - (b)->tv_usec;  
		if ((result)->tv_usec < 0) 
		{ 
			--(result)->tv_sec;
			(result)->tv_usec += 1000000; 
		}  
	} while (0); 

}
struct eventlist *events;
void insertevent(struct eventlist *insertme);
struct eventlist *event_find(struct eventlist *needle);

/* student-called routines */

event fish_scheduleevent(	int msec_delay,
							void (*event_handler)(void *),
							void *event_handler_argument) 
{
  struct timeval now;
  struct eventlist *newevent =  (struct eventlist *) malloc(sizeof(struct eventlist));

  SYSTEMTIME systemTime;
  GetSystemTime( &systemTime );

  SysTimeToTimeVal( &systemTime, &now );

 
  // printf("scheduling\n");
  
  newevent->when.tv_sec = now.tv_sec + (msec_delay / 1000);
  newevent->when.tv_usec = now.tv_usec + ((msec_delay * 1000) % 1000000);
  while(newevent->when.tv_usec > 1000000) 
  {
    newevent->when.tv_usec -= 1000000;
    newevent->when.tv_sec += 1;
  }
  newevent->event_handler = event_handler;
  newevent->event_handler_argument = event_handler_argument;

  if(event_find(newevent)) 
  {
    errorabort("Scheduled duplicate event, you should probably cancel\n"
               "the other one. Generating core dump for gdb.\n");
  } else 
  {
    insertevent(newevent);
  }
  
  return((event)newevent);
}


struct eventlist *event_find(struct eventlist *needle) 
{
  /* continue through the event list while
   *  a) there's more to the list, and 
   *  b) we haven't found the node we seek
   */
  struct eventlist *p;
  for(	p=events; 
		p != NULL && 
        ( p->event_handler != needle->event_handler || 
        p->event_handler_argument != needle->event_handler_argument);
		p=p->next);

  return(p);
}


void insertevent(struct eventlist *insertme) 
{
  struct eventlist *p, *q;
  /* find the last event before the new one */
  for(	p=events, q=NULL; 
		p != NULL &&
        timercmp(&p->when, &insertme->when, <=);
		q=p, p=p->next);

  if(q==NULL) 
  {
    insertme->next = events;
    events = insertme;
  } 
  else 
  {
    insertme->next = p;
    q->next = insertme;
  }
}

void *removeevent(struct eventlist *removeme) 
{
  struct eventlist *p, *q;
  void *retval = NULL;
  if(removeme == NULL) 
  { 
    errorabort( "removing an invalid event\n" );
  }
  /* find the last event before the new one */
  for(p=events, q=NULL; 
      p != NULL && p != removeme;
      q=p, p=p->next);
  if(p==removeme) 
  {
    if(q==NULL) 
	{
      events = p->next;
    } 
	else 
	{
      q->next = p->next;
    }
    retval = p->event_handler_argument;
    free(p);
  }

  if(p == NULL) 
  {
    // Commented to make life's easy: Sushant
    // errorabort("didn't find the event to remove\n");
    return NULL;
    
  }
  return(retval);
}

/* returns the event_handler_argument associated 
   with the event; you may free the data (unlikely) 
   or ignore the return value (more likely).  NULL
   is returned if the event was not found */
void *fish_cancelevent(int event_handle) 
{
  return(removeevent((struct eventlist *)event_handle));
}

void fish_executepending(void) 
{
  struct timeval now;
  struct eventlist *p;
  
  SYSTEMTIME systemTime;
  GetSystemTime( &systemTime );

  SysTimeToTimeVal( &systemTime, &now );

  
  for(p = events;
      p != NULL &&
        timercmp(&p->when, &now, <);
      p = events) 
  {
    /* huh; we have something to do. */
    events = p->next;
    (p->event_handler)(p->event_handler_argument);
    free(p);
  }
}

/* returns whether timeout should be set at all. */
int fish_time_to_next_event(struct timeval *delta_t) 
{
  struct timeval now;
  
  SYSTEMTIME systemTime;
  GetSystemTime( &systemTime );

  SysTimeToTimeVal( &systemTime, &now );

  if(events) 
  {
    if( timercmp(&events->when, &now, <) ) 
	{
      /* it's too late. return 0 time. */
      delta_t->tv_sec = 0;
      delta_t->tv_usec = 0;
    } 
	else 
	{
      /* defined in sys/time.h */
      timersub(&events->when, &now, delta_t);
    }
    return 1;
  } 
  else 
  {
    /* timeout should not be set at all, but just in case... */
    delta_t->tv_sec = -1;
    delta_t->tv_usec = -1;
    return 0;
  }
} 

/* necessary for a full interface, but not for student use */
void *fish_eventdata(event event_handle) 
{
  return ( ((struct eventlist *)event_handle) -> event_handler_argument );
}
